"""
This module provides helper functions for Google Pub/Sub.
"""

import base64
import json
import os


def decode(data, b64decode=True):
    """
    Decodes data coming from Pub/Sub.

    Args:
        data (str): The data you want to decode.
        b64decode (bool, optional): Whether to base64 decode the data.

    Returns:
        dict: The decoded message.
    """
    decoded = data
    if b64decode:
        decoded = base64.b64decode(decoded)
    return json.loads(decoded.decode("utf-8"))


def _encode(data, b64encode=False):
    encoded = json.dumps(data).encode("ascii")
    if b64encode:
        encoded = base64.b64encode(encoded)
    return encoded


def to_topic(messages, topic=None, b64encode=False):
    """
    Publishes messages to a Pub/Sub topic.

    Args:
        messages (list[dict]): The messages.
        topic (str, optional): The topic. If not provided, defaults to
            `os.environ["GCF_UTILS_PUBSUB_TOPIC"]`.
        b64encode (bool, optional): Whether to base64 encode the messages.
            This is done automatically for Google Cloud Functions.
    """
    # pylint: disable=import-outside-toplevel
    from google.cloud import pubsub_v1

    # pylint: enable=import-outside-toplevel

    publisher = pubsub_v1.PublisherClient()
    topic = topic or os.environ["GCF_UTILS_PUBSUB_TOPIC"]

    futures = [publisher.publish(topic, _encode(data, b64encode)) for data in messages]
    _ = [future.result() for future in futures]



"""
This module provides a decorator class to report errors to Google Cloud.
"""

from google.cloud.error_reporting import Client


# pylint: disable=invalid-name
class error_reporting:
    # pylint: enable=invalid-name
    """
    This class serves as a decorator to report errors to Google Cloud.

    Args:
        func (callable): The function you want to decorate.
    """

    def __init__(self, func):
        self._func = func
        self._client = Client()

    def __call__(self, *args, **kwargs):
        try:
            self._func(*args, **kwargs)
        except Exception as error:
            self._client.report_exception()
            raise error



"""
Helper functions for Google Cloud Services.
"""

from itertools import islice
import re

from .bigquery import to_table
from .error_reporting import error_reporting
from .pub_sub import to_topic, decode
from .storage import to_bucket


def chunks(iterable, chunksize=500):
    """
    Breaks down an iterable into chunks.
    """
    iterable = iter(iterable)
    while True:
        chunk = tuple(islice(iterable, chunksize))
        if not chunk:
            return
        yield chunk


def camel_to_snake(term):
    """
    Converts a CamedCased term into a snake_cased term.
    """
    pattern = re.compile(r"(?<!^)(?=[A-Z])")
    return pattern.sub("_", term).lower()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cloud_functions_utils']

package_data = \
{'': ['*']}

install_requires = \
['google-cloud-bigquery>=1.24.0,<2.0.0',
 'google-cloud-error-reporting>=0.33.0,<0.34.0',
 'google-cloud-pubsub>=1.4.3,<2.0.0',
 'google-cloud-storage>=1.28.0,<2.0.0']

setup_kwargs = {
    'name': 'cloud-functions-utils',
    'version': '0.2.4',
    'description': 'Utilities for Google Cloud Functions.',
    'long_description': None,
    'author': 'Jan-Benedikt Jagusch',
    'author_email': 'jan.jagusch@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
This module provides a function to insert rows into a Google BigQuery table.
"""

from datetime import datetime
import os

from google.cloud import bigquery

from .exceptions import EmptyRowsError, InsertionError


def _add_inserted_at_field(rows):
    inserted_at = datetime.now()
    for row in rows:
        row.update({"inserted_at": inserted_at})


def to_table(rows, project=None, dataset=None, table=None, add_inserted_at=True):
    """
    Inserts rows into Google BigQuery.

    Args:
        rows (list[dict]): The rows.
        project (str, optional): The name of the Google Cloud project. If not provided,
            defaults to `os.environ["GCF_UTILS_PROJECT"]`.
        dataset (str, optional): The name of the Google BigQuery dataset. If not
            provided, defaults to `os.environ["GCF_UTILS_BIGQUERY_DATASET"]`.
        table (str, optional): The name of the Google BigQuery table. If not provided,
            defaults to `os.environ["GCF_UTILS_BIGQUERY_TABLE"]`.
        add_inserted_at (bool, optional): If true, adds an `inserted_at` field with
            the current datetime to all rows.

    Raises:
        EmptyRowsError: When they rows object is an empty list.
        InsertionError: When an error occurs while inserting rows to BigQuery.
    """

    if not rows:
        raise EmptyRowsError

    project = project or os.environ["GCF_UTILS_PROJECT"]
    dataset = dataset or os.environ["GCF_UTILS_BIGQUERY_DATASET"]
    table = table or os.environ["GCF_UTILS_BIGQUERY_TABLE"]

    bq_client = bigquery.Client()
    bq_dataset = bigquery.dataset.DatasetReference.from_string(f"{project}.{dataset}")
    bq_table = bq_dataset.table(table)

    if add_inserted_at:
        _add_inserted_at_field(rows)

    errors = bq_client.insert_rows(bq_client.get_table(bq_table), rows)
    if errors:
        raise InsertionError(errors)



"""
Custom exceptions for this package.
"""


class CloudFunctionsUtilsError(Exception):
    """
    Abstract parent exception for this package.
    """


class InsertionError(CloudFunctionsUtilsError):
    """
    Error that is thrown when insertion into Google BigQuery does not succeed.
    """


class EmptyRowsError(CloudFunctionsUtilsError):
    """
    Error that is thrown when trying to insert an empty list to Google BigQuery.
    """



"""
This module provides a function to upload an object string to Google Cloud Storage.
"""

import os

from google.cloud import storage


def to_bucket(obj_str, filepath, bucket=None):
    """
    Uploads an object string to a filepath in a Google Cloud Storage bucket.

    Args:
        obj_str (str): The object string.
        filepath (str): The filepath.
        bucket (str, optional): The bucket. If not provided, defaults to
            `os.environ["GCF_UTILS_STORAGE_BUCKET"]`.
    """
    bucket = bucket or os.environ["GCF_UTILS_STORAGE_BUCKET"]

    storage_client = storage.Client()
    storage_bucket = storage_client.bucket(bucket)
    storage_blob = storage_bucket.blob(filepath)

    storage_blob.upload_from_string(obj_str)
